local cql = ClassicQuestLog

-- indexed by campaignID, an ordered table of lore for the campaign, if any
local campaignLore = {}

local campaignChapters = {} -- indexed by campaignID, cache of chapterIDs for the campaign
local campaignInfo = {} -- indexed by campaignID, cache of GetCampaignInfo
local questLines = {} -- indexed by chapterID, the questIDs in the chapter

local currentCampaignID -- currently viewed campaignID in lore panel

-- pools for regions that build out the scrollable lore content
local fontstringPool = {}
local texturePool = {}

function cql.Lore:Setup()
    self:SetScript("OnSizeChanged",self.OnSizeChanged)
    self:OnSizeChanged(self:GetWidth(),self:GetHeight())
end

function cql.Lore:OnSizeChanged(width,height)
    local content = self.ScrollFrame.Content
    local contentWidth = max(floor(width+0.5)-35,150)
    content:SetWidth(contentWidth)
    content.width = contentWidth

    self:Update() -- necessary for lore elements to adjust
end

-- requests the lore for all campaignIDs in the log
local requestedIDs = {} -- lookup table of campaignIDs requested
function cql.Lore:RequestCampaignLore()
    local found = false
    for i=1,C_QuestLog.GetNumQuestLogEntries() do
        local info = C_QuestLog.GetInfo(i)
        if info and not info.isHidden and info.isHeader and info.campaignID and not requestedIDs[info.campaignID] then
            C_LoreText.RequestLoreTextForCampaignID(info.campaignID)
            requestedIDs[info.campaignID] = true
            found = true
        end
    end
    wipe(requestedIDs)
end

-- processes the payload of LORE_TEXT_UPDATED_CAMPAIGN to update campaignLore for a single campaignID
-- presumably no lore ever gets removed in a session (ignoring events with 0 lore in payload)
function cql.Lore:UpdateCampaignLore(campaignID,lore)
    local somethingChanged = false
    if campaignID and type(lore)=="table" and #lore>0 then
        if not campaignLore[campaignID] then -- this campaignID had no lore
            campaignLore[campaignID] = lore
            somethingChanged = true
        elseif #campaignLore[campaignID]~=#lore then -- this campaignID's lore has changed
            campaignLore[campaignID] = lore
            somethingChanged = true
        end
    end
    -- if new lore added or changed, then update UI to update the lore buttons and lore window if it's up
    if somethingChanged and cql:IsVisible() then
        cql.timer:Start(0.25,cql.DelayedUpdate)
    end
end

-- returns true if the campaignID has campaignLore
function cql.Lore:HasLore(campaignID)
    return (campaignID and campaignLore[campaignID]) and true or false
end

-- changes to the lore panel for the given campaignID if it has lore
function cql.Lore:ShowCampaignID(campaignID)
    if campaignID and campaignLore[campaignID] then
        currentCampaignID = campaignID
        cql:SetMode("Lore")
    end
end

-- returns the lore being displayed
function cql.Lore:GetCurrentCampaignID()
    return currentCampaignID
end

function cql.Lore:GetCampaignChapterIDs(campaignID)
    if campaignID then
        if not campaignChapters[campaignID] then
            campaignChapters[campaignID] = C_CampaignInfo.GetChapterIDs(campaignID)
        end
        return campaignChapters[campaignID]
    end
end

-- returns a 0/0 Chapters string for the given campaignID
function cql.Lore:GetCampaignProgress(campaignID,formatOverride)
    if campaignID then
        local totalChapters = 0
        local completedChapters = 0
        local chapterIDs = self:GetCampaignChapterIDs(campaignID)

        if chapterIDs then
            for _,chapterID in ipairs(chapterIDs) do
                totalChapters = totalChapters + 1
                if C_QuestLine.IsComplete(chapterID) then
                    completedChapters = completedChapters + 1
                end
            end
        end

        return format(formatOverride or STORY_CHAPTERS,completedChapters,totalChapters)
    end
end

-- returns the C_CampaignInfo.GetCampaignInfo for a campaign and caches it
-- (these results should never change)
function cql.Lore:GetCampaignInfo(campaignID)
    if campaignID then
        if not campaignInfo[campaignID] then
            campaignInfo[campaignID] = C_CampaignInfo.GetCampaignInfo(campaignID)
        end
        return campaignInfo[campaignID]
    end
end

-- returns an ordered list of questIDs for the given chapterID
function cql.Lore:GetQuestLine(chapterID)
    if chapterID then
        if not questLines[chapterID] then
            questLines[chapterID]  = C_QuestLine.GetQuestLineQuests(chapterID)
        end
        return questLines[chapterID]
    end
end

-- returns numInProgress,numCompleted,numTotal for the given chapterID
function cql.Lore:GetChapterProgress(chapterID)
    if chapterID then
        local inProgress,completed,total = 0,0,0
        for _,questID in ipairs(self:GetQuestLine(chapterID)) do
            total = total + 1
            if C_QuestLog.IsOnQuest(questID) then
                inProgress = inProgress + 1
            end
            if C_QuestLog.IsQuestFlaggedCompleted(questID) then
                completed = completed + 1
            end
        end
        return inProgress,completed,total
    end
end

function cql.Lore:AcquireFontString()
    for _,fontstring in ipairs(fontstringPool) do
        if not fontstring.used then
            fontstring:ClearAllPoints()
            fontstring.used = true
            return fontstring
        end
    end
    -- if we reached here, no free fontstring, make a new one
    local fontstring = self.ScrollFrame.Content:CreateFontString(nil,"ARTWORK","GameFontNormal")
    fontstring.used = true
    tinsert(fontstringPool,fontstring)
    return fontstring
end

function cql.Lore:AcquireTexture()
    for _,texture in ipairs(texturePool) do
        if not texture.used then
            texture:ClearAllPoints()
            texture.used = true
            return texture
        end
    end
    -- if we reached here, no free texture, make a new one
    local texture = self.ScrollFrame.Content:CreateTexture(nil,"ARTWORK")
    texture.used = true
    --texture:SetHeight(16)
    texture:SetAtlas("Campaign-QuestLog-LoreDivider",true)
    tinsert(texturePool,texture)
    return texture
end

function cql.Lore:ReleasePools()
    for _,fontstring in ipairs(fontstringPool) do
        fontstring.used = nil
        fontstring:ClearAllPoints()
        fontstring:Hide()
    end
    for _,texture in ipairs(texturePool) do
        texture.used = nil
        texture:ClearAllPoints()
        texture:Hide()
    end
end

-- updates the lore frame with the currentCampaignID, or switches back to Detail
-- if the currentCampaignID is not valid (use ShowCampaignID to change campaigns)
function cql.Lore:Update()

    if not self:IsVisible() then
        return -- don't touch anything and leave if lore not on screen
    end

    local content = self.ScrollFrame.Content
    local campaignID = currentCampaignID
    local campaignInfo = self:GetCampaignInfo(campaignID)

    if not campaignID or not campaignLore[campaignID] or not campaignInfo then
        cql:SetMode("Detail")
        return -- if lore on screen but no known campaign, go back to details and leave
    end

    if cql.settings.HighContrast then
        self.Back:SetAtlas("QuestDetailsBackgrounds-Accessibility",false)
        self.Back:SetVertexColor(0.15,0.1,0.05)
        self.HeaderBackground:SetVertexColor(0.5,0.5,0.5)
    else
        self.Back:SetAtlas("Campaign-QuestLog-LoreBackground",false)
        self.Back:SetVertexColor(1,1,1)
        self.HeaderBackground:SetVertexColor(1,1,1)
    end

    --  campaign header stuff which spans whole Lore pane (content excludes scrollbar and padding)
    self.HeaderBackground:SetAtlas(campaignInfo.uiTextureKit and "Campaign_"..campaignInfo.uiTextureKit or "Campaign_Dracthyrawaken",false)
    self.Title:SetWidth(floor(self:GetWidth()-40+0.5))
    self.Title:SetText(campaignInfo.name)
    self.Progress:SetText(self:GetCampaignProgress(campaignID,CAMPAIGN_PROGRESS_CHAPTERS))
    self.Progress:SetShown(self.Title:GetStringHeight()<25)

    local width = content.width-24
    local lastObject

    self:ReleasePools()
    for index,info in ipairs(campaignLore[campaignID]) do
        local fontstring = self:AcquireFontString()
        if lastObject then
            fontstring:SetPoint("TOPLEFT",lastObject,"BOTTOMLEFT",0,-12)
        else
            fontstring:SetPoint("TOPLEFT",16,0)
        end
        fontstring:SetWidth(width)
        if info.isHeader then
            fontstring:SetFontObject(Game13FontShadow)
            fontstring:SetTextColor(HIGHLIGHT_FONT_COLOR:GetRGBA())
            fontstring:SetJustifyH("CENTER")
            fontstring:SetText("\n"..info.text)
            local texture = self:AcquireTexture()
            texture:SetPoint("BOTTOM",fontstring,"BOTTOM",0,-6)
            texture:SetWidth(width)
            texture:Show()
        else
            fontstring:SetFontObject(SystemFont_Shadow_Med1)
            fontstring:SetTextColor(LORE_TEXT_BODY_COLOR:GetRGBA())
            fontstring:SetJustifyH("LEFT")
            fontstring:SetText(info.text)
        end
        fontstring:Show()
        lastObject = fontstring
    end

end


function cql.Lore:ShowCampaignTooltip(campaignID,anchorPoint,relativeTo,relativePoint,xoff,yoff)

    if not campaignID or not anchorPoint or not relativeTo then
        return
    end

	GameTooltip:ClearAllPoints()
	GameTooltip:SetPoint(anchorPoint,relativeTo,relativePoint,xoff,yoff)
	GameTooltip:SetOwner(relativeTo,"ANCHOR_PRESERVE")

    local campaignInfo = self:GetCampaignInfo(campaignID)
    local chapterIDs = self:GetCampaignChapterIDs(campaignID)

    GameTooltip:SetText(format("\124cffffffff%s",campaignInfo.name))
    GameTooltip:AddLine(cql.Lore:GetCampaignProgress(campaignID,CAMPAIGN_PROGRESS_CHAPTERS))

    for index,chapterID in ipairs(chapterIDs) do
        if index==1 then -- if there's a chapterID to show, add an empty line to pad it
            GameTooltip:AddLine(" ")
        end
        local info = C_CampaignInfo.GetCampaignChapterInfo(chapterID)
        local inProgress,completed,total = cql.Lore:GetChapterProgress(chapterID)
        local text
        if inProgress>0 or completed>0 then
            if completed==total then
                GameTooltip:AddLine(format("%s (100%%)",info.name),0,0.9,0,true)
            else
                local progress = floor(((completed or 0)*100/(total or 1))+0.5)
                GameTooltip:AddLine(format("%s (%d%%)",info.name,progress),0.9,0.9,0.9,true)
            end
        else
            GameTooltip:AddLine(format("%s",info.name),0.55,0.55,0.55,true)
        end
    end

    GameTooltip:Show()
end